// Generated by CoffeeScript 2.3.1
(function() {
  // First-line handler for outbound calls
  // -------------------------------------
  var debug, pkg;

  pkg = require('../../../package.json');

  this.name = `${pkg.name}:middleware:client:egress:pre`;

  debug = (require('tangible'))(this.name);

  this.include = async function() {
    var base, base1, base2, base3, base4, base5, base6, base7, location, music_uri, number_domain, privacy, ref, ref1, ref2, ref3, src_number;
    if (((ref = this.session) != null ? ref.direction : void 0) !== 'egress') {
      return;
    }
    music_uri = (doc) => {
      if (doc.music == null) {
        return null;
      }
      return this.prompt.uri('prov', 'prov', doc._id, doc.music);
    };
    // Endpoint
    // --------
    // * session.endpoint_name (string) The name of the calling endpoint in an egress call, set from hdr.X-En
    this.session.endpoint_name = this.req.header('X-En');
    // Endpoint might be provided in the reference data for example for an `originate` call (in exultant-songs).
    if ((base = this.session).endpoint_name == null) {
      base.endpoint_name = (await this.reference.get_endpoint());
    }
    if (this.session.endpoint_name == null) {
      debug.dev('Missing endpoint_name', this.call.data);
      return this.respond('485 Missing Endpoint Header');
    }
    debug('endpoint', this.session.endpoint_name);
    await this.reference.set_endpoint(this.session.endpoint_name);
    await this.unset('sip_h_X-En');
    // * session.endpoint (object) Data from the calling `doc.endpoint` (also known as the `doc.src_endpoint`) in an egress call.
    this.session.endpoint = (await this.cfg.prov.get(`endpoint:${this.session.endpoint_name}`));
    await this.reference.set_endpoint(this.session.endpoint_name);
    if (this.session.endpoint.timezone != null) {
      this.session.timezone = this.session.endpoint.timezone;
    }
    if (this.session.endpoint.music != null) {
      this.session.music = music_uri(this.session.endpoint);
    }
    if (this.session.endpoint.trace) {
      this.session.dev_logger = true;
    }
    // If the call was blind-transfered by the agent, it will appear in a `transfer` direction, as coming from the agent (e.g. `sip_h_X-En` is copied over) even though the call actually connects the original caller and the recipient of the transfer. The call-id is the call-id of the original call (before the transfer). In this case, we should _not_ bind that call-id and the agent, since the agent is no longer part of that call.
    if (!this.session.transfer) {
      this.session.agent = this.session.endpoint_name;
      this.session.agent_name = (ref1 = this.session.endpoint.display_name) != null ? ref1 : this.source;
    }
    this.set({
      ccnq_endpoint: this.session.endpoint_name,
      ccnq_endpoint_json: JSON.stringify(this.session.endpoint)
    });
    // Outbound-route
    // --------------
    this.session.outbound_route = this.session.endpoint.outbound_route;
    if (this.session.outbound_route == null) {
      debug.dev('Missing outbound_route', this.session.endpoint);
      return this.respond('500 Endpoint has no outbound_route');
    }
    debug('outbound_route', this.session.outbound_route);
    // Number-domain
    // -------------

    // Note: The header is no longer generated by ccnq4-opensips.

    // * doc.number_domain On egress calls, the number-domain is selected based on the doc.src_endpoint.number_domain field.
    // * doc.src_endpoint.number_domain (required)
    number_domain = this.session.endpoint.number_domain;
    if (number_domain == null) {
      debug.dev('Missing number_domain', this.session.endpoint);
      return this.respond('480 Missing Number Domain');
    }
    // * session.number_domain Name of the number-domain (egress calls).
    this.session.number_domain = number_domain;
    await this.reference.set_number_domain(number_domain);
    // The `number_domain` field is required, but the number-domain record is optional.

    // * session.number_domain_data Data record of the number-domain (egress calls).
    this.session.number_domain_data = (await this.cfg.prov.get(`number_domain:${number_domain}`).catch((error) => {
      debug(`number_domain ${number_domain}: ${error}`);
      return {};
    }));
    debug('number_domain', number_domain);
    await this.reference.set_number_domain(number_domain);
    await this.user_tags(this.session.number_domain_data.tags);
    // Number-domain is less specific than endpoint, so do not override.
    if (this.session.number_domain_data.timezone != null) {
      if ((base1 = this.session).timezone == null) {
        base1.timezone = this.session.number_domain_data.timezone;
      }
    }
    if (this.session.number_domain_data.music != null) {
      if ((base2 = this.session).music == null) {
        base2.music = music_uri(this.session.number_domain_data);
      }
    }
    // Source (calling) number
    // -----------------------

    // * session.number (object) Data from doc.number for the calling number in an egress call.
    src_number = `${this.source}@${number_domain}`;
    if (this.session.transfer) {
      this.session.number = {};
    } else {
      this.session.number = (await this.cfg.prov.get(`number:${      // On a static trunk, the number might not be present.
src_number}`).catch((error) => {
        debug.ops(`number:${src_number}: ${error}`);
        return {};
      }));
    }
    if (this.session.number._id != null) {
      await this.reference.set_number(src_number);
    }
    // Number is more specific than endpoint or number-domain, override.
    if (this.session.number.timezone != null) {
      this.session.timezone = this.session.number.timezone;
    }
    if (this.session.number.music != null) {
      this.session.music = music_uri(this.session.number);
    }
    if (this.session.number.trace) {
      this.session.dev_logger = true;
    }
    // Dialplan
    // --------

    // * doc.local_number.dialplan (string) Dialplan used for egress calls. Generally `national` or `centrex`.
    // * doc.src_endpoint.dialplan (string) Dialplan used for egress calls if doc.local_number.dialplan is empty.
    // * doc.number_domain.dialplan (string) Dialplan used for egress calls if doc.local_number.dialplan and doc.src_endpoint.dialplan is empty.
    this.session.dialplan = this.session.number.dialplan;
    if ((base3 = this.session).dialplan == null) {
      base3.dialplan = this.session.endpoint.dialplan;
    }
    if ((base4 = this.session).dialplan == null) {
      base4.dialplan = this.session.number_domain_data.dialplan;
    }
    // Country
    // -------

    // * doc.local_number.country (string) Country used for egress calls.
    // * doc.src_endpoint.country (string) Country used for egress calls if doc.local_number.country is empty.
    // * doc.number_domain.country (string) Country used for egress calls if doc.local_number.country and doc.src_endpoint.country is empty.
    this.session.country = this.session.number.country;
    if ((base5 = this.session).country == null) {
      base5.country = this.session.endpoint.country;
    }
    if ((base6 = this.session).country == null) {
      base6.country = this.session.number_domain_data.country;
    }
    // Contrarily to established practices, our code uses lowercase country names.
    if (this.session.country != null) {
      this.session.country = this.session.country.toLowerCase();
    }
    debug('Ready', {
      endpoint_name: this.session.endpoint_name,
      outbound_route: this.session.outbound_route,
      number_domain: this.session.number_domain,
      dialplan: this.session.dialplan,
      country: this.session.country
    });
    // Location
    // --------

    // This is needed for emergency call routing.

    // * doc.local_number.location (string) Emergency call routing location.
    // * doc.src_endpoint.location (string) Emergency call routing location (if no doc.local_number.location is provided).
    location = this.session.number.location;
    if (location == null) {
      location = this.session.endpoint.location;
    }
    if (location == null) {
      location = '';
    }
    this.session.emergency_location = location;
    // Privacy
    // -------

    // Enforce configurable privacy settings.

    // * doc.local_number.privacy (boolean) Whether privacy is requested for egress calls.
    // * doc.src_endpoint.privacy (boolean) Whether privacy is requested for egress calls (if doc.local_number.privacy is not set).
    privacy = this.session.number.privacy;
    if (privacy == null) {
      privacy = this.session.endpoint.privacy;
    }
    if (privacy) {
      await this.action('privacy', 'number');
    } else {
      await this.action('privacy', 'no');
    }
    // Asserted-Number
    // ---------------

    // Enforce configurable Caller-ID. (Used in particular for ported-in numbers.)

    // * doc.local_number.asserted_number (string) Asserted number for egress calls.
    // * doc.src_endpoint.asserted_number (string) Asserted number for egress calls (if doc.local_number.asserted_number is not set).
    this.session.asserted = this.session.number.asserted_number;
    if ((base7 = this.session).asserted == null) {
      base7.asserted = this.session.endpoint.asserted_number;
    }
    // Check from
    // ----------

    // Optionally enforce that the calling number originates from the associated endpoint (useful e.g. to prevent invalid caller-id from static endpoints).

    // * doc.src_endpoint.check_from (boolean) If true, doc.local_number.endpoint must match the originating endpoint's name. This is used to restrict originating numbers on an endpoint to those that are provisioned on that endpoint.
    if (this.session.endpoint.check_from) {
      if (this.session.number.endpoint !== this.session.endpoint_name) {
        debug('From Username is not listed');
        return this.respond('403 From Username is not listed');
      }
    }
    // ICE
    // ---

    // For backward-compatibility we currently ignore ICE proposals.
    await this.set({
      ignore_sdp_ice: (ref2 = this.session.endpoint.ignore_sdp_ice) != null ? ref2 : true
    });
    // Call-recording
    // --------------
    if (this.session.number.record_egress) {
      this.record_call(this.session.number._id);
    }
    this.report({
      state: 'egress-received',
      endpoint: this.session.endpoint._id,
      number_domain: this.session.number_domain_data._id,
      number: (ref3 = this.session.number._id) != null ? ref3 : null
    });
    debug('OK');
  };

}).call(this);
